// @ts-ignore
const {$, $a, doAfterLoad, footer, header, nav} = window.fwdekker;

import {csrfToken, getApi, postApi, sharedMessageElement} from "./API";
import {CustomEventHandler} from "./CustomEventHandler";
import {clearMessage, clearMessages, showError, showSuccess, showWarning} from "./Message";


/**
 * Handles listeners to be invoked when the user logs in.
 */
const loginHandler = new CustomEventHandler();
/**
 * Handles listeners to be invoked when the user logs out.
 */
const logoutHandler = new CustomEventHandler();


/**
 * Refreshes the list of trackings in the table.
 */
function refreshTrackings(): void {
    getApi(
        {action: "list-trackings"},
        sharedMessageElement,
        (response) => {
            // Remove old rows
            $a("#trackings tbody tr:not(:last-child)").forEach((it: HTMLTableRowElement) => it.remove());

            // Insert new rows
            const tableBody = $("#trackings tbody");
            const lastRow = $("#trackings tbody tr:last-child");
            response.payload.forEach((tracking: any) => {
                const row = document.createElement("tr");

                const nameCell = document.createElement("td");
                const nameLink = document.createElement("a");
                nameLink.href = "https://en.wikipedia.org/wiki/" + tracking.name;
                nameLink.innerText = tracking.name;
                if (tracking.deleted)
                    nameLink.classList.add("redLink");
                nameCell.append(nameLink);
                row.append(nameCell);

                const statusCell = document.createElement("td");
                statusCell.innerText =
                    tracking.deleted
                        ? "page not found"
                        : (tracking.name === "Adolf Hitler" ? "deceased 🥳" : tracking.status);
                row.append(statusCell);

                const deleteCell = document.createElement("td");
                const deleteForm = document.createElement("form");
                deleteForm.addEventListener("submit", (event: SubmitEvent) => {
                    event.preventDefault();

                    postApi(
                        {action: "remove-tracking", token: csrfToken, person_name: tracking.name},
                        deleteForm,
                        () => refreshTrackings()
                    )
                });
                const deleteButton = document.createElement("button");
                deleteButton.innerText = "remove";
                deleteForm.append(deleteButton);
                deleteCell.append(deleteForm);
                row.append(deleteCell);

                tableBody.insertBefore(row, lastRow);
            });
        }
    );
}

/**
 * Refreshes displays of the user's data.
 */
function refreshUserData(): void {
    getApi(
        {action: "get-user-data"},
        sharedMessageElement,
        (response) => {
            const userData = response.payload;

            // Email
            $("#emailCurrent").innerText = userData.email;
            $("#emailVerified").innerText = userData.email_is_verified ? "yes" : "no";
            if (!userData.email_is_verified) {
                showWarning(
                    sharedMessageElement,
                    "You will not receive any email notifications until you verify your email address. " +
                    "Check your inbox for further instructions."
                );
                $("#resendEmailVerificationButton").classList.remove("hidden");
            }

            // Password update time
            const today = new Date();
            today.setHours(0, 0, 0, 0)
            const updateTime = new Date(userData.password_update_time * 1000);
            updateTime.setHours(0, 0, 0, 0);
            const diff = (+today - +updateTime) / 86400000;
            $("#passwordLastChanged").innerText = diff === 0 ? "today" : diff + " days ago";
        }
    )
}


// Initialize template
doAfterLoad(() => {
    $("#nav").appendChild(nav("/Tools/Death-Notifier/"));
    $("#header").appendChild(header({
        title: "Death Notifier",
        description: "Get notified when a famous person dies"
    }));
    $("#footer").appendChild(footer({
        vcsURL: "https://git.fwdekker.com/tools/death-notifier/",
        version: "v%%VERSION_NUMBER%%"
    }));
    $("main").classList.remove("hidden");
});

// Register event handlers
doAfterLoad(() => {
    loginHandler.addListener(() => {
        refreshUserData();
        refreshTrackings();

        $("#loginRow").classList.add("hidden")
        $("#trackingRow").classList.remove("hidden");
        $("#accountRow").classList.remove("hidden");
    });
    logoutHandler.addListener(() => {
        $("#loginRow").classList.remove("hidden")
        $("#trackingRow").classList.add("hidden");
        $("#accountRow").classList.add("hidden");
    });

    // Add event handlers
    const loginForm = $("#loginForm");
    loginForm.addEventListener("submit", (event: SubmitEvent) => {
        event.preventDefault();

        postApi(
            {
                action: "login",
                token: csrfToken,
                email: $("#loginEmail").value,
                password: $("#loginPassword").value,
            },
            event.target as HTMLFormElement,
            () => loginHandler.invokeListeners()
        );
    });
    loginHandler.addListener(() => {
        loginForm.reset();
        clearMessages(loginForm);
    });

    const registerForm = $("#registerForm");
    registerForm.addEventListener("submit", (event: SubmitEvent) => {
        event.preventDefault();

        postApi(
            {
                action: "register",
                token: csrfToken,
                email: $("#registerEmail").value,
                password: $("#registerPassword").value,
                password_confirm: $("#registerPasswordConfirm").value,
            },
            registerForm,
            () => {
                // TODO: Add client-side form validation
                registerForm.reset();
                showSuccess(
                    $(".formValidationInfo", registerForm),
                    "Account created successfully! You may now log in."
                );
            }
        );
    });
    loginHandler.addListener(() => {
        registerForm.reset();
        clearMessages(registerForm);
    });

    const logoutForm = $("#logoutForm");
    logoutForm.addEventListener("submit", (event: SubmitEvent) => {
        event.preventDefault();

        postApi(
            {
                action: "logout",
                token: csrfToken,
            },
            logoutForm,
            () => {
                logoutForm.reset();
                logoutHandler.invokeListeners();
            }
        );
    });

    const updateEmailForm = $("#updateEmailForm");
    updateEmailForm.addEventListener("submit", (event: SubmitEvent) => {
        event.preventDefault();

        postApi(
            {
                action: "update-email",
                token: csrfToken,
                email: $("#updateEmailEmail").value,
            },
            updateEmailForm,
            () => {
                updateEmailForm.reset();
                refreshUserData();
                showSuccess(
                    $(".formValidationInfo", updateEmailForm),
                    "Email address updated successfully! " +
                    "Check your inbox for the verification email. " +
                    "You will not receive notifications until you verify your email address."
                );
            }
        );
    });
    logoutHandler.addListener(() => {
        updateEmailForm.reset();
        clearMessages(updateEmailForm);
    });

    const resendEmailVerificationForm = $("#resendEmailVerificationForm");
    resendEmailVerificationForm.addEventListener("submit", (event: SubmitEvent) => {
        event.preventDefault();

        postApi(
            {action: "resend-verify-email", token: csrfToken},
            resendEmailVerificationForm,
            () => {
                resendEmailVerificationForm.reset();
                refreshUserData();
                showSuccess(
                    $(".formValidationInfo", resendEmailVerificationForm),
                    "Email verification resent successfully!"
                );
            }
        );
    });
    logoutHandler.addListener(() => {
        resendEmailVerificationForm.reset();
        clearMessages(resendEmailVerificationForm);
    });

    const updatePasswordForm = $("#updatePasswordForm");
    updatePasswordForm.addEventListener("submit", (event: SubmitEvent) => {
        event.preventDefault();

        postApi(
            {
                action: "update-password",
                token: csrfToken,
                password_old: $("#updatePasswordPasswordOld").value,
                password_new: $("#updatePasswordPasswordNew").value,
                password_confirm: $("#updatePasswordPasswordConfirm").value,
            },
            event.target as HTMLFormElement,
            () => {
                updatePasswordForm.reset();
                refreshUserData();
                showSuccess($(".formValidationInfo", updatePasswordForm), "Password updated successfully!");
            }
        );
    });
    logoutHandler.addListener(() => {
        updatePasswordForm.reset();
        clearMessages(updatePasswordForm);
    });

    const addTrackingForm = $("#addTrackingForm");
    addTrackingForm.addEventListener("submit", (event: SubmitEvent) => {
        event.preventDefault();

        postApi(
            {
                action: "add-tracking",
                token: csrfToken,
                person_name: $("#addTrackingPersonName").value,
            },
            addTrackingForm,
            () => {
                addTrackingForm.reset();
                refreshTrackings();
            }
        );
    });
    logoutHandler.addListener(() => {
        addTrackingForm.reset();
        clearMessages(addTrackingForm);
    });
});

// Run initialization code
doAfterLoad(() => {
    const get_params = new URLSearchParams(window.location.search);
    if (get_params.get("action") === "verify-email" && get_params.has("email") && get_params.has("token")) {
        postApi(
            {action: "verify-email", email: get_params.get("email"), token: get_params.get("token")},
            sharedMessageElement,
            () => {
                let secondsUntilRedirect = 3;
                const updateMessage = () => {
                    showSuccess(
                        sharedMessageElement,
                        `Your email address has been verified. ` +
                        `You will be redirected after ${secondsUntilRedirect} seconds.`
                    );
                    secondsUntilRedirect -= 1;
                    setTimeout(updateMessage, 1000);
                };
                updateMessage();

                setTimeout(() => window.location.href = "//" + location.host + location.pathname, 3000);
            },
            () => {
                $("#sharedHomeLink").classList.remove("hidden");

                showError(
                    sharedMessageElement,
                    "Failed to verify email address. Maybe you already verified your email address?"
                );
            }
        );
        return;
    }

    // Show content depending on whether user is logged in
    getApi(
        {action: "start-session"},
        sharedMessageElement,
        () => loginHandler.invokeListeners(),
        () => logoutHandler.invokeListeners()
    );
});
